/*
  zBoy - A GameBoy emulator for Windows and Linux
  Copyright (C) Mateusz Viste 2010, 2011

  ----------------------------------------------------------------------
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
  ----------------------------------------------------------------------
*/


#ifdef __zboy4linux__
  #pragma message("Compiling zBoy for Linux...")
#endif

#ifdef __zboy4windows__
  #pragma message("Compiling zBoy for Windows...")
  /* #define printf(...) */ /* Ignore any printf() calls on windows */
  /* void printf(char *s, ...) {
  } */
#endif


/* include some standard C headers */
#include <pthread.h>    /* for multi-threading capabilities */
#include <stdio.h>      /* for printf() and fprintf() */
#include <stdlib.h>     /* for atoi() and exit(), random() and randomize() */
#include <time.h>       /* for time-related functions (clock, time...) */
#include <stdint.h>     /* for uint8_t type */
#include <signal.h>
#include "SDL/SDL.h"
#include <getopt.h>

#define pVer "0.51"
#define pDate "2010,2011,2012"

/*  The options below make zBoy to be compiled in debug mode, and optionally trace CPU activity to a trace file */
/* #define DEBUGMODE */
/* #define DEBUGTRACEFILE */
/* #define MEMDUMP_AT_EXIT */

/* Produce a stand-alone exe file with an embedded ROM file */
/* #define EMBEDROM "embedrom/dracula.c" */
/* #define EMBEDROM "embedrom/pacman.c" */
/* #define EMBEDROM "embedrom/drmario.c" */

#ifndef DEBUGMODE   /* If not in debugmode, then disable PrintDebug */
  /* #define PrintDebug printf */
  /* #define PrintDebug(...) */
  void PrintDebug(char *s, ...) {
  }
#endif

struct RomInformations {
  char Title[256];
  char NiceTitle[256];       /* Same as title, but with lower/upper chars, spaces etc. */
  char Filename[1024];       /* The filename of the ROM (with full path) */
  char ShortFilename[1024];  /* The filename of the ROM (without path) */
  char ShortFilenameNoExt[1024];  /* The filename of the ROM (without path and without extension) */
  signed char ColorGB;       /* 0 = not color, 1 = color (but still GB compatible), 2 = GBC only (not GB compatible) */
  signed char SuperGB;       /* 0 = GameBoy, 1 = Super GameBoy */
  char CartTypeString[256];  /* Descriptive string for the cart type */
  unsigned int CrcSum;
  signed char MbcModel;      /* 0 = ROM ONLY  1 = MBC1  2 = MBC2... -1 = unknown */
  signed char Rumble;        /* 0 = No rumble  1 = Rumble pak (found only on some MBC5 chips) */
  signed char TimerRTC;      /* 0 = No RTC  1 = RTC  (found only on some MBC3 chips) */
  signed char Battery;       /* 0 = no battery, 1 = battery included */
  signed char Destination;   /* 0 = Japan, 1 = Other */
  char Licensee[256];
  int RomSize;
  int RamSize;
};

enum refreshscreenalgorithm {
  REFRESHSCREEN_NOSCALE,
  REFRESHSCREEN_BASICSCALE,
  REFRESHSCREEN_2X,
  REFRESHSCREEN_3X,
  REFRESHSCREEN_EAGLE
};

struct zboyparamstype {
  uint8_t ShowFPS;
  uint8_t NoScreenSaver;
  uint8_t ShowRomInfos;
  uint8_t NoSpeedLimit;
  uint8_t HiScoresMem;
  uint8_t NoCpuIdle;
  uint8_t GraphicScaleFactor;
  uint8_t fpslimit;
  char windowtitle[256]; /* used to hold the window's title */
  enum refreshscreenalgorithm scalingalgo;
};

struct RomInformations RomInfos;

char CartTypes[256][256];
signed char QuitEmulator;
signed int AskForScreenshot = 0;
unsigned int LastScreenshotTime = 0;
uint8_t NetPlay = 0;
char *NetPlayPeer;
char *NetPlayBindaddr = NULL;

unsigned int GraphicWidth, GraphicHeight;
char UserMessage[1024];
unsigned int UserMessageFramesLeft = 0;

#include "declares.h"         /* Some SUB/FUNCTION declarations */
#include "portable.c"         /* Various portable routines */
#include "binary.h"           /* Provides basic support to binary notation (eg. bx01100111) */
#include "adjtiming.c"        /* AdjustTiming() */
#include "numconv.c"          /* UbyteToByte() and DwordVal() */
#include "presskey.c"         /* PressAnyKey() */
#include "getsavedir.c"       /* Provides the GetSaveDir() call */
#include "getsnapshotfilename.c"
#include "mmu.c"              /* Include Memory Management Unit emulation core */
#include "hiscores.c"         /* Provides SaveHiScoresMem() and LoadHiScoresMem() */
#include "video.c"            /* Video emulation core */
#include "loadpal.c"          /* LoadPalette() */
#include "joypad.c"           /* Joypad interface emulation */
#include "cpu-timer.c"        /* Include timer emulation */
#include "cpu-divider.c"      /* Include divider emulation */
#include "cpu-z80.c"          /* Z80 CPU emulation core */
#ifdef DEBUGMODE
  #include "debug.c"          /* Includes some debugging routines */
#endif
#include "battery.c"          /* Provides LoadBattRAM() and SaveBattRAM() */
#include "font.c"             /* Includes font data used by zBoy to display onscreen messages */
#include "wordwrap.c"         /* WordWrap(), required by PrintMsg() */
#include "printmsg.c"         /* PrintMsg(text AS STRING) */
#include "loadrom.c"          /* LoadROM(RomFile AS STRING) */
#include "loadrominfos.c"     /* LoadRomInfos() */
#include "serial.c"           /* Serial interface emulation */
#include "savepcxfile.c"      /* Screenshot routine SavePcxFile() */
#include "save.c"             /* SaveGame() and LoadGame() */
#include "reset.c"            /* ResetEmulator() */
#include "ssaver.c"           /* InhibitScreenSaver() */
#include "about.c"            /* printhelp() */




int main (int argc, char *argv[]) {
  char *NoYes[2] = {"No", "Yes"};
  struct zboyparamstype zboyparams;
  int x,option_index;
  int UsedCycles;
  char c;

  static struct option long_options[] =
             {
               /* These options set a flag. */
               {"bootlogo", no_argument, 0, 'b'},
               {"eaglescale", required_argument, 0, 'e'},
               {"fpslimit", required_argument, 0, 'f'},
               {"scale", required_argument, 0, 's'},
               {"scale2x", required_argument, 0, 'S'},
               {"scale3x", required_argument, 0, 'c'},
               {"palette", required_argument, 0, 'p'},
               {"input", required_argument, 0, 'i'},
               {"nocpuidle", no_argument, 0, 'n'},
               {"nospeedlimit", no_argument, 0, 'o'},
               {"showfps", no_argument, 0, 'w'},
               {"noscreensaver", no_argument, 0, 'N'},
               {"hiscoresmem", no_argument, 0, 'H'},
               {"rominfos", no_argument, 0, 'r'},
               {"rom", required_argument, 0, 'R'},
               {"netplay", required_argument, 0, 't'},
               {"bindaddr", required_argument, 0, 'd'},
               {"help", no_argument, 0, 'h'},
               {0, 0, 0, 0}
             };

  /* Prepare SDL-related variables */
  SDL_Surface *screen;      /* virtual SDL screen area */
  SDL_Joystick *joystick;   /* joystick handler */
  SDL_Color palcolors[256]; /* SDL palette */
  /* double LastDebugOnscreen, OneSecondPollingTimer, LastScreenSaverPoke; */
  pthread_t thread_InhibitScreenSaver;

  /* Prepare thread attribute for threads that we'll launch later */
  pthread_attr_t threadattr;                                         /* This is *mandatory* to run */
  pthread_attr_init(&threadattr);                                    /* threads as detached. Otherwise */
  pthread_attr_setdetachstate(&threadattr, PTHREAD_CREATE_DETACHED); /* you're going for massive mem leaks. */
  
  /* Set zBoy parameters to default values */
  zboyparams.ShowRomInfos = 0;
  zboyparams.NoScreenSaver = 0;
  zboyparams.ShowFPS = 0;
  zboyparams.NoSpeedLimit = 0;
  zboyparams.HiScoresMem = 0;
  zboyparams.NoCpuIdle = 0;
  zboyparams.scalingalgo = REFRESHSCREEN_2X;  /* By default, use Scale2x scaling... */
  zboyparams.GraphicScaleFactor = 4;          /* ...with 4x magnification. */
  zboyparams.fpslimit = 60;

  /* Here I am preloading a default configuration for input keys mapping */
  if (SetKeys("keyboard") != 0) {
    printf("Error: Invalid input keys definitions.\n");
    return -1;
  }

  /*printf("Parsing command-line parameters...\n"); */

  while( (c = getopt_long(argc, argv, "bhnowNHre:f:s:S:c:p:i:R:t:d:", long_options, &option_index) ) != -1)
	switch(c){
		case 'h' : printhelp();
			   return(0);
		case 's' : if ((atoi(optarg) < 1) || (atoi(optarg) > 8)) {
              RomInfos.Filename[0] = 0;  /* Clear out filename, to make sure to display usage screen later */
     	  		   } else if (atoi(optarg) == 1) {
              zboyparams.scalingalgo = REFRESHSCREEN_NOSCALE; /* No scaling at all */
            } else {
				zboyparams.GraphicScaleFactor = atoi(optarg);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_BASICSCALE; /* Nearest Neighbor scaling */
          }
			   break;
		case 'S' : if ((atoi(optarg) != 2) && (atoi(optarg) != 4) && (atoi(optarg) != 6) && (atoi(optarg) != 8)) {
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
            } else {
				zboyparams.GraphicScaleFactor = atoi(optarg);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_2X;
		           }break;
		case 'c' : if ((atoi(optarg) != 3) && (atoi(optarg) != 6)) {
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
            } else {
				zboyparams.GraphicScaleFactor = atoi(optarg);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_3X; /* aka AdvMAME3x */
          }
			  break;
		case 'e' : if ((atoi(optarg) != 2) && (atoi(optarg) != 4) && (atoi(optarg) != 6) && (atoi(optarg) != 8)) {
              RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
            } else {
				zboyparams.GraphicScaleFactor = atoi(optarg);  /* retrieve provided scale parameter */
              zboyparams.scalingalgo = REFRESHSCREEN_EAGLE;
          }
			   break;
		case 'f' : if ((atoi(optarg) < 1) || (atoi(optarg) > 60)) {
            RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
          } else {
				zboyparams.fpslimit = atoi(optarg);  /* retrieve provided skipframes parameter */
          }
			   break;
		case 'b' : BootRomEnabledFlag = 1;
			   break;
		case 'p' : if (LoadPalette(optarg) != 0) {
             RomInfos.Filename[0] = 0;   /* Clear out filename, to make sure to display usage screen later */
          }
			   break;
		case 'r' : zboyparams.ShowRomInfos = 1;
			   break;
		case 'n' : zboyparams.NoCpuIdle = 1;
			   break;
		case 'o' : zboyparams.NoSpeedLimit = 1;
			   break;
		case 'N' : zboyparams.NoScreenSaver = 1; 
			   break;  
		case 'H' : zboyparams.HiScoresMem = 1; 
			   break;
		case 't' : NetPlay = 1; /* atoi(argv[x]); */
			   NetPlayPeer = optarg;
			   break;
		case 'd' : NetPlayBindaddr = optarg;
			   break;
		case 'R' : snprintf(RomInfos.Filename, 1022, optarg);
			   break;
		case 'i' : if (SetKeys(optarg) != 0)
          RomInfos.Filename[0] = 0;
			   break;
        }



  #ifdef DEBUGMODE
    GraphicWidth = 800
    GraphicHeight = 900
    GraphicScaleFactor = 2
  #else
    GraphicWidth = 160 * zboyparams.GraphicScaleFactor;
    GraphicHeight = 144 * zboyparams.GraphicScaleFactor;
  #endif

  /* Initialize the SDL library... */
  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
    printf("Couldn't initialize SDL: %s\n", SDL_GetError());
    return(1);
  }
  SDL_ShowCursor(SDL_DISABLE); /* Disable (hide) mouse cursor */
  atexit(SDL_Quit); /* ...and remember to deallocate SDL before quitting. */
  joystick = SDL_JoystickOpen(0);  /* ...and open the 1st system joystick */

  /* if (FullscreenFlag != 0) {
      switch (GraphicScaleFactor) {
        case 1:
          GraphicWidth = 320;
          GraphicHeight = 200;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 27;    // add 27 pixels of border on the top of the screen to vertically center the display area
          break;
        case 2:
          GraphicWidth = 512;
          GraphicHeight = 384;
          ScreenBorderRight = 96;  // add 96 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 47;    // add 47 pixels of border on the top of the screen to vertically center the display area
          break;
        case 3:
          GraphicWidth = 640;
          GraphicHeight = 480;
          GraphicScaleFactor = 3;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 23;    // add 24 pixels of border on the top of the screen to vertically center the display area
          break;
        case 4:
          GraphicWidth = 800;
          GraphicHeight = 600;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 11;    // add 11 pixels of border on the top of the screen to vertically center the display area
          break;
        case 5:
          GraphicWidth = 1024;
          GraphicHeight = 768;
          ScreenBorderRight = 112; // add 112 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 23;    // add 23 pixels of border on the top of the screen to vertically center the display area
          break;
        case 6:
          GraphicWidth = 1280;
          GraphicHeight = 1024;
          ScreenBorderRight = 160; // add 160 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 80;    // add 80 pixels of border on the top of the screen to vertically center the display area
          break;
        case 7:
          GraphicWidth = 1280;
          GraphicHeight = 1024;
          ScreenBorderRight = 80;  // add 80 pixels of border on the right side of the screen to horizontally center the display area
          ScreenBorderTop = 8;     // add 23 pixels of border on the top of the screen to vertically center the display area
          break;
      }

      // Set up the fullscreen video mode (some available flags: SDL_SWSURFACE, SDL_DOUBLEBUF, SDL_HWSURFACE, SDL_FULLSCREEN...)
      screen = SDL_SetVideoMode(GraphicWidth, GraphicHeight, 8, SDL_SWSURFACE);
      if (screen == NULL) {
        printf("Couldn't set video mode: %s\n", SDL_GetError());
        return(1);
      }
    } else { */
      /* Set up a video mode (that is, create a window on screen) */
      screen = SDL_SetVideoMode(GraphicWidth, GraphicHeight, 8, SDL_SWSURFACE);
      if (screen == NULL) {
        printf("Couldn't set video mode: %s\n", SDL_GetError());
        return(1);
      }
  /* } */
  SDL_WM_SetCaption("zBoy", "zBoy");
  /* On Windows, use a dirty trick to hide that ugly title-bar icon... */
  #ifdef __zboy4windows__
  do {
    SDL_Surface *dummyicon;
    dummyicon = SDL_CreateRGBSurface(SDL_SWSURFACE, 32, 32, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0xFFFFFFFF);
    SDL_WM_SetIcon(dummyicon, NULL);
  } while(0);
  #endif

  InitScreenOldBuffer();   /* Init the screen buff used for LCD change detection (will force the complete draw of the 1st frame) */

  printf("Uploading palette into the virtual screen...\n");
  for (x = 0; x < 256; x++) {
    palcolors[x].r = (ScreenPalette32[x] >> 16) & 0xFF;
    palcolors[x].g = (ScreenPalette32[x] >> 8) & 0xFF;
    palcolors[x].b = ScreenPalette32[x] & 0xFF;
  }
  SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, palcolors, 0, 255);

  #ifndef EMBEDROM
    if (RomInfos.Filename[0] == 0) {
      printf("No ROM filename provided!\n");
      sprintf(UserMessage, "\n  *** ZBOY V%s ***\n\nUSAGE: ZBOY FILE.GB\n\nSEE ZBOY.TXT FOR DETAILED INSTRUCTIONS.", pVer);
      PrintMsg(UserMessage, 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
    }
    printf("Loading the ROM data into memory...\n");
    x = LoadRom(&RomInfos);
    if (x != 0) {
      printf("Error: Could not load the ROM file. [%d]\n", x);
      switch (x) {
        case -1:
          PrintMsg("ERROR: THE FILE IS TOO SMALL TO BE A VALID GAMEBOY ROM!", 0);
          RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
          PressAnyKey();
          return(0);
          break;
        case -2:
          PrintMsg("ERROR: THE FILE IS TOO BIG TO BE A VALID GAMEBOY ROM!", 0);
          RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
          PressAnyKey();
          return(0);
          break;
        case -3:
          PrintMsg("ERROR: THE SPECIFIED ROM FILE COULD NOT BE OPENED!", 0);
          RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
          PressAnyKey();
          return(0);
          break;
        case -4:
          PrintMsg("ERROR: THE FILE IS NOT A VALID GAMEBOY ROM!", 0);
          RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
          PressAnyKey();
          return(0);
          break;
        default:
          PrintMsg("ERROR: UNEXPECTED ERROR OCCURED WHILE LOADING THE ROM!", 0);
          RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
          PressAnyKey();
          return(0);
          break;
      }
    }
  #endif

  if (NetPlay != 0) {
    if ((s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) printf("socket problem\n");
    memset((char *) &si_me, 0, sizeof(si_me));
    si_me.sin_family = AF_INET;
    /* if (NetPlay == 1) si_me.sin_port = htons(8765); else si_me.sin_port = htons(8764); */
    si_me.sin_port = htons(8764);
    if (NetPlayBindaddr == NULL) {
        si_me.sin_addr.s_addr = htonl(INADDR_ANY); /* listen on all addresses if no bind specified */
      } else {
        if (inet_aton(NetPlayBindaddr, &si_me.sin_addr) == 0) printf("inet_aton() failed - bindaddr is probably not a valid IP\n");
    }
    memset((char *) &si_other, 0, sizeof(si_other));
    si_other.sin_family = AF_INET;
    /* if (NetPlay == 1) si_other.sin_port = htons(8764); else si_other.sin_port = htons(8765); */
    si_other.sin_port = htons(8764);
    /* if (inet_aton("127.0.0.1", &si_other.sin_addr)==0) printf("inet_aton() failed - NetPlay peer is probably not a valid IP\n"); */
    if (inet_aton(NetPlayPeer, &si_other.sin_addr) == 0) printf("inet_aton() failed - NetPlay peer is probably not a valid IP\n");
    if (bind(s, (struct sockaddr *) &si_me, sizeof(si_me)) == -1) printf("bind() problem\n");
  }

  printf("Loading ROM informations...\n");
  LoadRomInfos(&RomInfos);

  if (zboyparams.ShowRomInfos != 0) {
    char messagetoprint[1024];
    sprintf(messagetoprint, "TITLE: %s\n\nCRC32: %08X\n\nLICENSEE: %s\n\nCART TYPE: %s\n\nBATTERY: %s\n\nROM SIZE: %u KIB\n\nRAM SIZE: %u KIB\n\nPRESS ANY KEY...", RomInfos.NiceTitle, RomInfos.CrcSum, RomInfos.Licensee, RomInfos.CartTypeString, NoYes[RomInfos.Battery], RomInfos.RomSize >> 10, RomInfos.RamSize >> 10);
    /* Switch the string to all-uppercase */
    for (x = 0; messagetoprint[x] != 0; x++) if ((messagetoprint[x] >= 'a') && (messagetoprint[x] <= 'z')) messagetoprint[x] -= 32;
    PrintMsg(messagetoprint, 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    printf("Title......: %s (\"%s\")\n", RomInfos.Title, RomInfos.NiceTitle);
    printf("Filename...: %s\n", RomInfos.Filename);
    printf("CRC32 sum..: %8X\n", RomInfos.CrcSum);
    printf("Licensee...: %s\n", RomInfos.Licensee);
    if (RomInfos.ColorGB == 2) {
        printf("ColorGB....: Yes (GBC ONLY)\n");
      } else {
        printf("ColorGB....: %s\n", NoYes[RomInfos.ColorGB]);
    }
    printf("SuperGB....: %s\n", NoYes[RomInfos.SuperGB]);
    printf("Cart Type..: %s\n", RomInfos.CartTypeString);
    printf("Battery....: %s\n", NoYes[RomInfos.Battery]);
    if (RomInfos.Destination == 0) {
        printf("Country....: Japan\n");
      } else {
        printf("Country....: Other (not Japan)\n");
    }
    printf("ROM Size...: %u KiB\n", (RomInfos.RomSize >> 10));
    printf("RAM Size...: %u KiB\n", (RomInfos.RamSize >> 10));
    PressAnyKey();
  }

  if (RomInfos.Rumble != 0) {
    PrintMsg("THIS ROM USES A RUMBLE MECHANISM. ZBOY DOESN'T SUPPORT THAT YET! SORRY.", 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }
  if (RomInfos.TimerRTC != 0) {
    PrintMsg("THIS ROM USES AN INTERNAL RTC TIMER CLOCK. ZBOY DOESN'T SUPPORT SUCH ROMS YET! SORRY.", 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }
  /*
  IF RomInfos.SuperGB = 1 THEN PRINT : PRINT "Error: SuperGB ROMs are not supported (yet). Sorry." : SLEEP : END
  IF RomInfos.ColorGB = 2 THEN PRINT : PRINT "Error: This ROM requires a GameBoy Color emulator. zBoy supports only legacy GameBoy. Sorry." : SLEEP : END
  */

  switch (RomInfos.MbcModel) {  /* Check the MBC chip type, and select the right Read/Write functions */
    case 0:   /* ROM ONLY */
      MemoryWriteSpecial = &MBC0_MemoryWrite;
      MemoryReadSpecial = &MBC0_MemoryRead;
      break;
    case 1:   /* MBC1 */
      MemoryWriteSpecial = &MBC1_MemoryWrite;
      MemoryReadSpecial = &MBC1_MemoryRead;
      break;
    case 2:   /* MBC2 */
      MemoryWriteSpecial = &MBC2_MemoryWrite;
      MemoryReadSpecial = &MBC2_MemoryRead;
      break;
    case 3:   /* MBC3 */
      MemoryWriteSpecial = &MBC3_MemoryWrite;
      MemoryReadSpecial = &MBC3_MemoryRead;
      break;
    case 5:   /* MBC5 */
      MemoryWriteSpecial = &MBC5_MemoryWrite;
      MemoryReadSpecial = &MBC5_MemoryRead;
      break;
    default:  /* Any other (not supported) memory controller */
      PrintMsg("ERROR: THE MEMORY CONTROLLER USED BY THIS ROM IS NOT SUPPORTED (YET) BY ZBOY. SORRY.", 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
      break;
  }

  /* Set up the window's title */
  if (RomInfos.NiceTitle[0] != 0) {
      snprintf(zboyparams.windowtitle, 256, "%s (zBoy)", RomInfos.NiceTitle);
    } else {
      snprintf(zboyparams.windowtitle, 256, "zBoy");
  }
  SDL_WM_SetCaption(zboyparams.windowtitle, "zBoy");

  srand(time(NULL));  /* Seed the random generator */

  if (zboyparams.NoScreenSaver != 0) {  /* If ScreenSaver inhibition is turned on, launch the InhibitScreenSaver thread */
    if (pthread_create(&thread_InhibitScreenSaver, &threadattr, &InhibitScreenSaver, NULL) != 0) printf("Error: Inhibit ScreenSaver thread_create failed.\n");
  }

  /*InitRAM() */    /* Init RAM areas to random values (as in a real GameBoy) */
  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) LoadBattRAM();  /* check for battery-saved RAM */

  ResetEmulator(); /* Fire up the Zilog80 monster, and init all required variables */

  if (zboyparams.HiScoresMem != 0) LoadHiScoresMem();   /* Check out if need to look after hiscores */

  printf("Starting emulation...\n");

  do {
    UsedCycles = CpuExec();
    TotalCycles += UsedCycles;    /* Increment the global cycles counter */
    uTimer(UsedCycles);           /* Update uTimer */
    incDivider(UsedCycles);       /* Increment the DIV register */
    VideoSysUpdate(UsedCycles, screen, &zboyparams);   /* Update Video subsystem */

    CheckSerialLink(UsedCycles);  /* Check serial link (not supported yet) */
    CheckJoypad(UsedCycles, &zboyparams);  /* Update the Joypad register */
    CheckInterrupts();  /* Check if any interrupt has been requested since last time */
    #ifdef DEBUGMODE
      if ((TIMER - LastDebugOnscreen) > 0.5) {
        DebugSpriteDump(1);  /* perform some debug dumps onscreen (OBJ table at &h8000) */
        DebugSpriteDump(2);  /* perform some debug dumps onscreen (OBJ table at &h8800) */
        DebugPaletteDump();  /* Dump system's palette onscreen */
        DebugShowCPUspeed(); /* Calc current CPU speed and put it onscreen */
        DebugShowCpuRegisters(); /* Display state of CPU registers */
        DebugOnscreen();     /* Display debug buffer informations onscreen */
        LastDebugOnscreen = TIMER;
      }
    #endif
  } while (QuitEmulator == 0);

  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) SaveBattRAM();  /* check for battery-saved RAM */
  if (zboyparams.HiScoresMem != 0) SaveHiScoresMem();

  SDL_JoystickClose(joystick);

  #ifdef MEMDUMP_AT_EXIT
    DIM AS UINTEGER TempUINT             /* ----------------- */
    OPEN "memdump.bin" FOR OUTPUT AS #1  /* This is a memory  */
    FOR TempUINT = &h0000 TO &hFFFF      /* dump routine.     */
      MemoryRead(TempUINT, x)            /*     --  --        */
      PUT #1,, x                         /* Don't include in  */
    NEXT TempUINT                        /* official release! */
    CLOSE #1                             /* ----------------- */
  #endif

  #ifdef DEBUGMODE
    if (QuitEmulator != 0) {
      DebugOnscreen()
      SLEEP
    }
  #endif

  if (NetPlay != 0) close(s);

  #ifdef NETPLAYSUPPORT
    IF (NetPlay = 1) THEN  /* Close net stuff if NetPlay is used */
      NeedToCloseNetworkThreads = 1
      ThreadWait(NetPollingThread)
      ThreadWait(NetSendingThread)
    END IF
  #endif

  fflush(stdin); /* Flush keyb buffer */

  return(0);
}
